"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.isQueryValid = exports.parseQuery = void 0;
const utils_1 = require("../utils");
const parser_1 = require("./parser");
const parser = new parser_1.SoqlParser();
const BaseSoqlVisitor = parser.getBaseCstVisitorConstructor();
const BOOLEANS = ['TRUE', 'FALSE'];
const DATE_LITERALS = [
    'YESTERDAY',
    'TODAY',
    'TOMORROW',
    'LAST_WEEK',
    'THIS_WEEK',
    'NEXT_WEEK',
    'LAST_MONTH',
    'THIS_MONTH',
    'NEXT_MONTH',
    'LAST_90_DAYS',
    'NEXT_90_DAYS',
    'THIS_QUARTER',
    'LAST_QUARTER',
    'NEXT_QUARTER',
    'THIS_YEAR',
    'LAST_YEAR',
    'NEXT_YEAR',
    'THIS_FISCAL_QUARTER',
    'LAST_FISCAL_QUARTER',
    'NEXT_FISCAL_QUARTER',
    'THIS_FISCAL_YEAR',
    'LAST_FISCAL_YEAR',
    'NEXT_FISCAL_YEAR',
];
const DATE_N_LITERALS = [
    'NEXT_N_DAYS',
    'LAST_N_DAYS',
    'N_DAYS_AGO',
    'NEXT_N_WEEKS',
    'LAST_N_WEEKS',
    'N_WEEKS_AGO',
    'NEXT_N_MONTHS',
    'LAST_N_MONTHS',
    'N_MONTHS_AGO',
    'NEXT_N_QUARTERS',
    'LAST_N_QUARTERS',
    'N_QUARTERS_AGO',
    'NEXT_N_YEARS',
    'LAST_N_YEARS',
    'N_YEARS_AGO',
    'NEXT_N_FISCAL_QUARTERS',
    'LAST_N_FISCAL_QUARTERS',
    'N_FISCAL_QUARTERS_AGO',
    'NEXT_N_FISCAL_YEARS',
    'LAST_N_FISCAL_YEARS',
    'N_FISCAL_YEARS_AGO',
];
class SOQLVisitor extends BaseSoqlVisitor {
    constructor() {
        super();
        this.helpers = {
            $_getFieldFunction: (ctx, isAggregateFn = false, includeType = true) => {
                const args = ctx.functionExpression
                    ? ctx.functionExpression.map((node) => this.visit(ctx.functionExpression, { includeType })).flat()
                    : [];
                const output = {};
                if (includeType) {
                    output.type = 'FieldFunctionExpression';
                }
                output.functionName = ctx.fn[0].tokenType.name;
                output.parameters = args;
                if (includeType && isAggregateFn) {
                    output.isAggregateFn = isAggregateFn;
                }
                output.rawValue = `${ctx.fn[0].image}(${args.map((arg) => (typeof arg === 'string' ? arg : arg.rawValue)).join(', ')})`;
                return output;
            },
            $_getLiteralTypeFromTokenType: (tokenTypeName) => {
                if (tokenTypeName === 'REAL_NUMBER') {
                    return 'DECIMAL';
                }
                else if (tokenTypeName === 'CURRENCY_PREFIXED_DECIMAL') {
                    return 'DECIMAL_WITH_CURRENCY_PREFIX';
                }
                else if (tokenTypeName === 'CURRENCY_PREFIXED_INTEGER') {
                    return 'INTEGER_WITH_CURRENCY_PREFIX';
                }
                else if (tokenTypeName === 'SIGNED_DECIMAL') {
                    return 'DECIMAL';
                }
                else if (tokenTypeName === 'UNSIGNED_DECIMAL') {
                    return 'DECIMAL';
                }
                else if (tokenTypeName === 'UNSIGNED_INTEGER') {
                    return 'INTEGER';
                }
                else if (tokenTypeName === 'SIGNED_INTEGER') {
                    return 'INTEGER';
                }
                else if (tokenTypeName === 'DATETIME') {
                    return 'DATETIME';
                }
                else if (tokenTypeName === 'DATE') {
                    return 'DATE';
                }
                else if (tokenTypeName === 'NULL') {
                    return 'NULL';
                }
                else if (tokenTypeName === 'StringIdentifier') {
                    return 'STRING';
                }
                else if (tokenTypeName === 'Identifier') {
                    return 'STRING';
                }
                else if (BOOLEANS.includes(tokenTypeName)) {
                    return 'BOOLEAN';
                }
                else if (DATE_LITERALS.includes(tokenTypeName)) {
                    return 'DATE_LITERAL';
                }
                else if (DATE_N_LITERALS.includes(tokenTypeName)) {
                    return 'DATE_N_LITERAL';
                }
                else {
                    return 'STRING';
                }
            },
            /**
             * Shared logic since there are two entry points depending on if allowPartialQuery=true
             */
            $_parseSelect(ctx, options) {
                const { isSubquery } = options || { isSubquery: false };
                let output = {};
                if (ctx.selectClause) {
                    output.fields = this.visit(ctx.selectClause);
                }
                if (ctx.fromClause) {
                    if ((0, utils_1.isSubqueryFromFlag)(output, isSubquery)) {
                        const { sObject, alias, sObjectPrefix } = this.visit(ctx.fromClause);
                        output.relationshipName = sObject;
                        if (alias) {
                            output.sObjectAlias = alias;
                        }
                        if (sObjectPrefix) {
                            output.sObjectPrefix = sObjectPrefix;
                        }
                    }
                    else {
                        const { sObject, alias } = this.visit(ctx.fromClause);
                        output.sObject = sObject;
                        if (alias) {
                            output.sObjectAlias = alias;
                        }
                    }
                }
                if (Array.isArray(output.fields)) {
                    if (!!output.sObjectAlias) {
                        output.fields.forEach((field) => {
                            if (field.relationships && field.relationships[0] === output.sObjectAlias) {
                                field.relationships = field.relationships.slice(1);
                                field.objectPrefix = output.sObjectAlias;
                            }
                            if (field.relationships && field.relationships.length === 0) {
                                delete field.relationships;
                                field.type = 'Field';
                            }
                        });
                    }
                }
                output = Object.assign(Object.assign({}, output), this.visit(ctx.clauseStatements));
                return output;
            },
        };
        this.validateVisitor();
    }
    /**
     * Public entry point 1: `selectStatement`
     * @param ctx
     * @param options
     */
    selectStatement(ctx, options) {
        return this.helpers.$_parseSelect.bind(this)(ctx, options);
    }
    /**
     * Public entry point 2: `selectStatementPartial`
     * @param ctx
     * @param options
     */
    selectStatementPartial(ctx, options) {
        return this.helpers.$_parseSelect.bind(this)(ctx, options);
    }
    clauseStatements(ctx) {
        const query = {};
        if (ctx.usingScopeClause && !ctx.usingScopeClause[0].recoveredNode) {
            query.usingScope = this.visit(ctx.usingScopeClause);
        }
        if (ctx.whereClause && !ctx.whereClause[0].recoveredNode) {
            query.where = this.visit(ctx.whereClause);
        }
        if (ctx.withClause) {
            ctx.withClause
                .filter(item => !item.recoveredNode)
                .forEach(item => {
                const { withSecurityEnforced, withAccessLevel, withDataCategory } = this.visit(item);
                if (withSecurityEnforced) {
                    query.withSecurityEnforced = withSecurityEnforced;
                }
                if (withAccessLevel) {
                    query.withAccessLevel = withAccessLevel;
                }
                if (withDataCategory) {
                    query.withDataCategory = withDataCategory;
                }
            });
        }
        if (ctx.groupByClause && !ctx.groupByClause[0].recoveredNode) {
            query.groupBy = this.visit(ctx.groupByClause);
        }
        if (ctx.havingClause && !ctx.havingClause[0].recoveredNode) {
            query.having = this.visit(ctx.havingClause);
        }
        if (ctx.orderByClause && !ctx.orderByClause[0].recoveredNode) {
            query.orderBy = this.visit(ctx.orderByClause);
        }
        if (ctx.limitClause && !ctx.limitClause[0].recoveredNode) {
            query.limit = Number(this.visit(ctx.limitClause));
        }
        if (ctx.offsetClause && !ctx.offsetClause[0].recoveredNode) {
            query.offset = Number(this.visit(ctx.offsetClause));
        }
        if (ctx.forViewOrReference && !ctx.forViewOrReference[0].recoveredNode) {
            query.for = this.visit(ctx.forViewOrReference);
        }
        if (ctx.updateTrackingViewstat && !ctx.updateTrackingViewstat[0].recoveredNode) {
            query.update = this.visit(ctx.updateTrackingViewstat);
        }
        return query;
    }
    selectClause(ctx) {
        if (ctx.field) {
            return ctx.field.map(item => {
                if ((0, utils_1.isToken)(item)) {
                    const field = item.image;
                    let output;
                    if (!field.includes('.')) {
                        output = {
                            type: 'Field',
                            field: field,
                        };
                    }
                    else {
                        const splitFields = field.split('.');
                        output = {
                            type: 'FieldRelationship',
                            field: splitFields[splitFields.length - 1],
                            relationships: splitFields.slice(0, splitFields.length - 1),
                            rawValue: field,
                        };
                    }
                    return output;
                }
                else {
                    return this.visit(item);
                }
            });
        }
        return [];
    }
    selectClauseFunctionIdentifier(ctx) {
        let output = Object.assign({}, this.visit(ctx.fn));
        if (ctx.alias) {
            output.alias = ctx.alias[0].image;
        }
        return output;
    }
    selectClauseSubqueryIdentifier(ctx) {
        return {
            type: 'FieldSubquery',
            subquery: this.visit(ctx.selectStatement, { isSubquery: true }),
        };
    }
    selectClauseTypeOf(ctx) {
        let conditions = ctx.selectClauseTypeOfThen.map((item) => this.visit(item));
        if (ctx.selectClauseTypeOfElse) {
            conditions.push(this.visit(ctx.selectClauseTypeOfElse));
        }
        return {
            type: 'FieldTypeof',
            field: ctx.typeOfField[0].image,
            conditions,
        };
    }
    selectClauseIdentifier(ctx) {
        const item = ctx.field[0];
        const alias = !!ctx.alias ? ctx.alias[0].image : undefined;
        const field = item.image;
        let output;
        if (!field.includes('.')) {
            output = {
                type: 'Field',
                field: field,
            };
        }
        else {
            const splitFields = field.split('.');
            output = {
                type: 'FieldRelationship',
                field: splitFields[splitFields.length - 1],
                relationships: splitFields.slice(0, splitFields.length - 1),
                rawValue: field,
            };
        }
        if (alias) {
            output.alias = alias;
        }
        return output;
    }
    selectClauseTypeOfThen(ctx) {
        return {
            type: 'WHEN',
            objectType: ctx.typeOfField[0].image,
            fieldList: ctx.field.map((item) => item.image),
        };
    }
    selectClauseTypeOfElse(ctx) {
        return {
            type: 'ELSE',
            fieldList: ctx.field.map((item) => item.image),
        };
    }
    fromClause(ctx) {
        let sObject = ctx.Identifier[0].image;
        let output;
        if (sObject.includes('.')) {
            const sObjectPrefix = sObject.split('.');
            output = {
                sObjectPrefix: sObjectPrefix.slice(0, sObjectPrefix.length - 1),
                sObject: sObjectPrefix[sObjectPrefix.length - 1],
            };
        }
        else {
            output = {
                sObject,
            };
        }
        if (ctx.alias && ctx.alias[0]) {
            output.alias = ctx.alias[0].image;
        }
        return output;
    }
    usingScopeClause(ctx) {
        return ctx.UsingScopeEnumeration[0].image;
    }
    whereClauseSubqueryIdentifier(ctx) {
        return this.visit(ctx.selectStatement, { isSubquery: false });
    }
    whereClause(ctx) {
        const where = ctx.conditionExpression.reduce((expressions, currExpression) => {
            if (!expressions.expressionTree) {
                const tempExpression = this.visit(currExpression);
                expressions.expressionTree = tempExpression;
                expressions.prevExpression = tempExpression.right ? tempExpression.right : tempExpression;
            }
            else {
                const tempExpression = this.visit(currExpression, { prevExpression: expressions.prevExpression });
                expressions.prevExpression.right = tempExpression;
                expressions.prevExpression = tempExpression.right ? tempExpression.right : tempExpression;
            }
            return expressions;
        }, { prevExpression: undefined, expressionTree: undefined });
        return where.expressionTree;
    }
    conditionExpression(ctx, options) {
        options = options || {};
        if (options.prevExpression && ctx.logicalOperator) {
            options.prevExpression.operator = ctx.logicalOperator[0].tokenType.name;
        }
        let baseExpression = {};
        let currExpression = baseExpression;
        if (Array.isArray(ctx.expressionNegation)) {
            baseExpression = this.visit(ctx.expressionNegation);
            currExpression = baseExpression.right;
        }
        currExpression.left = this.visit(ctx.expression);
        return baseExpression;
    }
    withClause(ctx) {
        if (ctx.withSecurityEnforced) {
            return {
                withSecurityEnforced: true,
            };
        }
        else if (ctx.withAccessLevel) {
            return {
                withAccessLevel: ctx.withAccessLevel[0].image,
            };
        }
        else {
            return {
                withDataCategory: {
                    conditions: this.visit(ctx.withDataCategory),
                },
            };
        }
    }
    withDataCategory(ctx) {
        return ctx.withDataCategoryArr.map(item => this.visit(item));
    }
    withDataCategoryArr(ctx) {
        return {
            groupName: ctx.dataCategoryGroupName[0].image,
            selector: ctx.filteringSelector[0].image,
            parameters: ctx.dataCategoryName.map((item) => item.image),
        };
    }
    groupByClause(ctx) {
        return ctx.groupBy.map((groupBy) => ((0, utils_1.isToken)(groupBy) ? { field: groupBy.image } : { fn: this.visit(groupBy, { includeType: false }) }));
    }
    havingClause(ctx) {
        // expressionWithAggregateFunction
        const having = ctx.conditionExpression.reduce((expressions, currExpression) => {
            if (!expressions.expressionTree) {
                expressions.expressionTree = this.visit(currExpression);
                expressions.prevExpression = expressions.expressionTree;
            }
            else {
                expressions.prevExpression.right = this.visit(currExpression, {
                    prevExpression: expressions.prevExpression,
                });
                expressions.prevExpression = expressions.prevExpression.right;
            }
            return expressions;
        }, { prevExpression: undefined, expressionTree: undefined });
        return having.expressionTree;
    }
    orderByClause(ctx) {
        return ctx.orderByExpressionOrFn.map(item => this.visit(item));
    }
    orderByExpression(ctx) {
        const orderByClause = {
            field: ctx.Identifier[0].image,
        };
        if (ctx.order && ctx.order[0]) {
            orderByClause.order = ctx.order[0].tokenType.name;
        }
        if (ctx.nulls && ctx.nulls[0]) {
            orderByClause.nulls = ctx.nulls[0].tokenType.name;
        }
        return orderByClause;
    }
    orderByGroupingFunctionExpression(ctx) {
        const orderByClause = {
            fn: this.helpers.$_getFieldFunction(ctx, false, false),
        };
        if (ctx.order && ctx.order[0]) {
            orderByClause.order = ctx.order[0].tokenType.name;
        }
        if (ctx.nulls && ctx.nulls[0]) {
            orderByClause.nulls = ctx.nulls[0].tokenType.name;
        }
        return orderByClause;
    }
    orderBySpecialFunctionExpression(ctx) {
        const orderByClause = {};
        if (ctx.aggregateFunction) {
            orderByClause.fn = this.visit(ctx.aggregateFunction, { includeType: false });
        }
        else if (ctx.dateFunction) {
            orderByClause.fn = this.visit(ctx.dateFunction, { includeType: false });
        }
        else if (ctx.locationFunction) {
            orderByClause.fn = this.visit(ctx.locationFunction, { includeType: false });
        }
        if (ctx.order && ctx.order[0]) {
            orderByClause.order = ctx.order[0].tokenType.name;
        }
        if (ctx.nulls && ctx.nulls[0]) {
            orderByClause.nulls = ctx.nulls[0].tokenType.name;
        }
        return orderByClause;
    }
    limitClause(ctx) {
        return ctx.value[0].image;
    }
    offsetClause(ctx) {
        return ctx.value[0].image;
    }
    dateFunction(ctx, options = { includeType: true }) {
        return this.helpers.$_getFieldFunction(ctx, false, options.includeType);
    }
    aggregateFunction(ctx, options = { includeType: true }) {
        return this.helpers.$_getFieldFunction(ctx, true, options.includeType);
    }
    fieldsFunction(ctx, options = { includeType: true }) {
        let output = {};
        if (options.includeType) {
            output.type = 'FieldFunctionExpression';
        }
        output = Object.assign(Object.assign({}, output), {
            functionName: 'FIELDS',
            parameters: [ctx.params[0].image],
        });
        output.rawValue = `FIELDS(${output.parameters[0]})`;
        return output;
    }
    otherFunction(ctx, options = { includeType: true }) {
        return this.helpers.$_getFieldFunction(ctx, false, options.includeType);
    }
    cubeFunction(ctx) {
        return this.helpers.$_getFieldFunction(ctx, false, false);
    }
    rollupFunction(ctx) {
        return this.helpers.$_getFieldFunction(ctx, false, false);
    }
    locationFunction(ctx, options = { includeType: true }) {
        let output = {};
        if (options.includeType) {
            output.type = 'FieldFunctionExpression';
        }
        output = Object.assign(Object.assign({}, output), {
            functionName: 'DISTANCE',
            parameters: [
                ctx.location1[0].image,
                (0, utils_1.isToken)(ctx.location2) ? ctx.location2[0].image : this.visit(ctx.location2, options),
                ctx.unit[0].image,
            ],
        });
        if (options.includeType) {
            output.isAggregateFn = true;
        }
        output.rawValue = `DISTANCE(${output.parameters[0]}, ${(0, utils_1.isString)(output.parameters[1]) ? output.parameters[1] : output.parameters[1].rawValue}, ${output.parameters[2]})`;
        return output;
    }
    geolocationFunction(ctx, options = { includeType: true }) {
        let output = {};
        if (options.includeType) {
            output.type = 'FieldFunctionExpression';
        }
        output = Object.assign(Object.assign({}, output), {
            functionName: 'GEOLOCATION',
            parameters: [ctx.latitude[0].image, ctx.longitude[0].image],
            rawValue: `GEOLOCATION(${ctx.latitude[0].image}, ${ctx.longitude[0].image})`,
        });
        return output;
    }
    functionExpression(ctx, options = { includeType: true }) {
        if (ctx.params) {
            return ctx.params.map((item) => {
                if (item.image) {
                    return item.image;
                }
                return this.visit(item, options);
            });
        }
        return [];
    }
    expression(ctx) {
        // const { value, literalType, dateLiteralVariable } = this.visit(ctx.rhs, { returnLiteralType: true });
        const { value, literalType, dateLiteralVariable, operator } = this.visit(ctx.operator, { returnLiteralType: true });
        const output = {};
        if ((0, utils_1.isToken)(ctx.lhs)) {
            output.field = ctx.lhs[0].image;
        }
        else {
            output.fn = this.visit(ctx.lhs, { includeType: false });
        }
        // output.operator = this.visit(ctx.relationalOperator) || this.visit(ctx.setOperator);
        output.operator = operator;
        if (literalType === 'SUBQUERY') {
            output.valueQuery = value;
        }
        else {
            output.value = value;
            output.literalType = literalType;
        }
        if (dateLiteralVariable) {
            output.dateLiteralVariable = dateLiteralVariable;
        }
        if (ctx.L_PAREN) {
            output.openParen = ctx.L_PAREN.length;
        }
        if (ctx.R_PAREN) {
            output.closeParen = ctx.R_PAREN.length;
        }
        return output;
    }
    expressionPartWithNegation(ctx) {
        const output = {
            left: ctx.L_PAREN ? { openParen: ctx.L_PAREN.length } : null,
            operator: 'NOT',
            right: {
                left: {},
            },
        };
        return output;
    }
    expressionWithRelationalOperator(ctx) {
        return Object.assign({ operator: this.visit(ctx.relationalOperator) || this.visit(ctx.setOperator) }, this.visit(ctx.rhs, { returnLiteralType: true }));
    }
    expressionWithSetOperator(ctx) {
        return Object.assign({ operator: this.visit(ctx.relationalOperator) || this.visit(ctx.setOperator) }, this.visit(ctx.rhs, { returnLiteralType: true }));
    }
    atomicExpression(ctx, options) {
        options = options || {};
        let value;
        let literalType;
        let dateLiteralVariable;
        if (ctx.apexBindVariableExpression) {
            value = this.visit(ctx.apexBindVariableExpression);
            literalType = 'APEX_BIND_VARIABLE';
        }
        else if (ctx.NumberIdentifier) {
            value = ctx.NumberIdentifier[0].image;
            literalType = this.helpers.$_getLiteralTypeFromTokenType(ctx.NumberIdentifier[0].tokenType.name);
        }
        else if (ctx.UnsignedInteger) {
            value = ctx.UnsignedInteger[0].image;
            literalType = 'INTEGER';
        }
        else if (ctx.SignedInteger) {
            value = ctx.SignedInteger[0].image;
            literalType = 'INTEGER';
        }
        else if (ctx.RealNumber) {
            value = ctx.RealNumber[0].image;
            literalType = 'DECIMAL';
        }
        else if (ctx.DateIdentifier) {
            value = ctx.DateIdentifier[0].image;
            literalType = this.helpers.$_getLiteralTypeFromTokenType(ctx.DateIdentifier[0].tokenType.name);
        }
        else if (ctx.CurrencyPrefixedInteger) {
            value = ctx.CurrencyPrefixedInteger[0].image;
            literalType = 'INTEGER_WITH_CURRENCY_PREFIX';
        }
        else if (ctx.CurrencyPrefixedDecimal) {
            value = ctx.CurrencyPrefixedDecimal[0].image;
            literalType = 'DECIMAL_WITH_CURRENCY_PREFIX';
        }
        else if (ctx.DateTime) {
            value = ctx.DateTime[0].image;
            literalType = 'DATETIME';
        }
        else if (ctx.date) {
            value = ctx.DateToken[0].image;
            literalType = 'DATE';
        }
        else if (ctx.NULL) {
            value = 'NULL';
            literalType = 'NULL';
        }
        else if (ctx.StringIdentifier) {
            value = ctx.StringIdentifier[0].image;
            literalType = 'STRING';
        }
        else if (ctx.Identifier) {
            value = ctx.Identifier[0].image;
            literalType = 'STRING';
        }
        else if (ctx.booleanValue) {
            value = this.visit(ctx.booleanValue);
            literalType = 'BOOLEAN';
        }
        else if (ctx.DateLiteral) {
            value = ctx.DateLiteral[0].image;
            literalType = 'DATE_LITERAL';
        }
        else if (ctx.dateNLiteral) {
            const valueAndVariable = this.visit(ctx.dateNLiteral);
            value = valueAndVariable.value;
            dateLiteralVariable = valueAndVariable.variable;
            literalType = 'DATE_N_LITERAL';
        }
        else if (ctx.arrayExpression) {
            const arrayValues = this.visit(ctx.arrayExpression);
            value = arrayValues.map((item) => item.value);
            const dateLiteralTemp = arrayValues.map((item) => item.variable || null);
            const hasDateLiterals = dateLiteralTemp.some(item => item !== null);
            if (new Set(arrayValues.map((item) => item.type)).size === 1) {
                literalType = this.helpers.$_getLiteralTypeFromTokenType(arrayValues[0].type);
            }
            else {
                literalType = arrayValues.map((item) => this.helpers.$_getLiteralTypeFromTokenType(item.type));
            }
            if (hasDateLiterals) {
                dateLiteralVariable = dateLiteralTemp;
            }
            literalType = literalType || 'STRING';
        }
        else if (ctx.whereClauseSubqueryIdentifier) {
            value = this.visit(ctx.whereClauseSubqueryIdentifier);
            literalType = 'SUBQUERY';
        }
        if (options.returnLiteralType) {
            return {
                value,
                literalType,
                dateLiteralVariable,
            };
        }
        else {
            return value;
        }
    }
    apexBindVariableExpression(ctx) {
        return ctx.apex.map(item => this.visit(item)).join('.');
    }
    apexBindVariableIdentifier(ctx) {
        let output = ctx.Identifier[0].image;
        if (ctx.apexBindVariableFunctionArrayAccessor) {
            output += this.visit(ctx.apexBindVariableFunctionArrayAccessor[0]);
        }
        return output;
    }
    apexBindVariableNewInstantiation(ctx) {
        let output = `new ${ctx.function[0].image}`;
        if (ctx.apexBindVariableGeneric) {
            output += this.visit(ctx.apexBindVariableGeneric[0]);
        }
        output += this.visit(ctx.apexBindVariableFunctionParams[0]);
        if (ctx.apexBindVariableFunctionArrayAccessor) {
            output += this.visit(ctx.apexBindVariableFunctionArrayAccessor[0]);
        }
        return output;
    }
    apexBindVariableFunctionCall(ctx) {
        let output = `${ctx.function[0].image}${this.visit(ctx.apexBindVariableFunctionParams[0])}`;
        if (ctx.apexBindVariableFunctionArrayAccessor) {
            output += this.visit(ctx.apexBindVariableFunctionArrayAccessor[0]);
        }
        return output;
    }
    apexBindVariableGeneric(ctx) {
        return `<${ctx.parameter.map(item => item.image).join(', ')}>`;
    }
    apexBindVariableFunctionParams(ctx) {
        const params = Array.isArray(ctx.parameter) ? ctx.parameter : [];
        return `(${params.map(item => item.image).join(', ')})`;
    }
    apexBindVariableFunctionArrayAccessor(ctx) {
        return `[${ctx.value[0].image}]`;
    }
    arrayExpression(ctx) {
        return ctx.value.map((item) => {
            if ((0, utils_1.isToken)(item)) {
                return {
                    type: item.tokenType.name,
                    value: item.image,
                };
            }
            else {
                return this.visit(item, { includeType: true });
            }
        });
    }
    relationalOperator(ctx) {
        return ctx.operator[0].image;
    }
    setOperator(ctx) {
        return ctx.operator[0].tokenType.name.replace('_', ' ');
    }
    booleanValue(ctx) {
        return ctx.boolean[0].tokenType.name;
    }
    dateNLiteral(ctx, options) {
        const output = {
            value: `${ctx.dateNLiteral[0].image}:${ctx.variable[0].image}`,
            variable: Number(ctx.variable[0].image),
        };
        if (options && options.includeType) {
            output.type = ctx.dateNLiteral[0].tokenType.name;
        }
        return output;
    }
    forViewOrReference(ctx) {
        return ctx.value[0].tokenType.name;
    }
    updateTrackingViewstat(ctx) {
        return ctx.value[0].tokenType.name;
    }
}
// Our visitor has no state, so a single instance is sufficient.
const visitor = new SOQLVisitor();
/**
 * Parse soql query into Query data structure.
 *
 * @param soql
 * @param options
 */
function parseQuery(soql, options) {
    const { cst } = (0, parser_1.parse)(soql, options);
    const query = visitor.visit(cst);
    return query;
}
exports.parseQuery = parseQuery;
/**
 * Lex and parse query (without walking parsed results) to determine if query is valid.
 * options.ignoreParseErrors will not be honored
 * @param soql
 */
function isQueryValid(soql, options) {
    try {
        const { parseErrors } = (0, parser_1.parse)(soql, options);
        return parseErrors.length === 0 ? true : false;
    }
    catch (ex) {
        return false;
    }
}
exports.isQueryValid = isQueryValid;
